"""TKET is an open source quantum compiler developed by Quantinuum.

Central to TKET's design is its hardware agnosticism which allows researchers
and quantum software developers to take advantage of its powerful compilation
for many different quantum architectures.

This is the Python front-end for TKET, providing a high-level interface for
working with quantum circuits. See also the Rust library with the same name on
[crates.io](https://crates.io/crates/tket).
"""

from . import circuit, ops, optimiser, passes, pattern, rewrite

__all__ = ["circuit", "ops", "optimiser", "passes", "pattern", "rewrite"]


# This is updated by our release-please workflow, triggered by this
# annotation: x-release-please-version
__version__ = "0.12.0"



from pathlib import Path
from tket._tket.circuit import Node, Tk2Circuit

class ECCRewriter:
    @staticmethod
    def load_precompiled(filename: Path) -> ECCRewriter:
        """Load a precompiled rewriter from a file."""

    def get_rewrites(self, circ: Tk2Circuit) -> list[CircuitRewrite]:
        """Get rewrites for a circuit."""

class CircuitRewrite:
    """A rewrite rule for circuits."""

    def __init__(
        self,
        source_position: Subcircuit,
        source_circ: Tk2Circuit,
        replacement: Tk2Circuit,
    ) -> None:
        """Create a new circuit rewrite rule."""

    def node_count_delta(self) -> int:
        """The change in node count from the rewrite.

        A positive value indicates an increase in node count.
        """

    def replacement(self) -> Tk2Circuit:
        """The replacement circuit."""

class Subcircuit:
    """A subcircuit of a circuit."""

    def __init__(self, nodes: list[Node], circ: Tk2Circuit) -> None:
        """Create a new subcircuit."""



from pytket import Circuit
from tket.optimiser import BadgerOptimiser


def test_simple_optimiser():
    """a simple circuit matching test"""
    c = Circuit(3).CX(0, 1).CX(0, 1).CX(1, 2)
    opt = BadgerOptimiser.compile_eccs("test_files/cx_cx_eccs.json")

    # optimise for 1s
    cc = opt.optimise(c, 3)
    exp_c = Circuit(3).CX(1, 2)

    assert cc == exp_c



from pytket import Circuit, OpType
from dataclasses import dataclass
from typing import Callable, Any
from tket.passes import badger_pass, greedy_depth_reduce, chunks
from tket.circuit import Tk2Circuit
from tket.pattern import Rule, RuleMatcher
import hypothesis.strategies as st
from hypothesis.strategies._internal import SearchStrategy
from hypothesis import given, settings

import pytest


@st.composite
def circuits(
    draw: Callable[[SearchStrategy[Any]], Any],
    n_qubits: SearchStrategy[int] = st.integers(min_value=0, max_value=8),
    depth: SearchStrategy[int] = st.integers(min_value=5, max_value=50),
) -> Circuit:
    total_qubits = draw(n_qubits)
    circuit = Circuit(total_qubits)
    if total_qubits == 0:
        return circuit
    for _ in range(draw(depth)):
        gates = [circuit.Rz, circuit.H]
        if total_qubits > 1:
            gates.extend([circuit.CX])
        gate = draw(st.sampled_from(gates))
        control = draw(st.integers(min_value=0, max_value=total_qubits - 1))
        if gate in (circuit.CX,):
            target = draw(
                st.integers(min_value=0, max_value=total_qubits - 1).filter(
                    lambda x: x != control
                )
            )
            gate(control, target)
        if gate == circuit.Rz:
            angle = draw(st.floats(min_value=-2.0, max_value=2.0))
            gate(angle, control)
        if gate == circuit.H:
            gate(control)
    return circuit


@pytest.mark.skip(
    reason="bug to be investigated, see https://github.com/CQCL/tket2/issues/983"
)
def test_simple_badger_pass_no_opt():
    c = Circuit(3).CCX(0, 1, 2)
    badger = badger_pass(max_threads=1, timeout=0, rebase=True)
    badger.apply(c)
    assert c.n_gates_of_type(OpType.CX) == 6


@dataclass
class DepthOptimisePass:
    def apply(self, circ: Circuit) -> Circuit:
        (circ, n_moves) = greedy_depth_reduce(circ)
        return circ


def test_depth_optimise():
    c = Circuit(4).CX(0, 2).CX(1, 2).CX(1, 3)

    assert c.depth() == 3

    c = DepthOptimisePass().apply(c)

    assert c.depth() == 2


def _depth_impl(circ: Circuit) -> None:
    new, _ = greedy_depth_reduce(circ)

    assert circ.n_gates == new.n_gates
    assert new.depth() <= circ.depth()


@given(circ=circuits())
@settings(print_blob=True, deadline=30)
def test_depth_hyp(circ: Circuit) -> None:
    _depth_impl(circ)


def test_depth_bug() -> None:
    circ = Circuit(3).H(0).CX(1, 0).H(0).CX(0, 2).H(0).CX(1, 2)
    _depth_impl(circ)


def test_chunks():
    c = Circuit(4).CX(0, 2).CX(1, 3).CX(1, 2).CX(0, 3).CX(1, 3)

    assert c.depth() == 3

    circ_chunks = chunks(c, 2)
    circuits = circ_chunks.circuits()
    circ_chunks.update_circuit(0, circuits[0])
    c2 = circ_chunks.reassemble()

    assert c2.depth() == 3
    assert type(c2) is Circuit

    # Split and reassemble, with a tket circuit
    tk2_chunks = chunks(Tk2Circuit(c2), 2)
    tk2 = tk2_chunks.reassemble()

    assert type(tk2) is Tk2Circuit


def test_cx_rule():
    c = Tk2Circuit(Circuit(4).CX(0, 2).CX(1, 2).CX(1, 2))

    rule = Rule(Circuit(2).CX(0, 1).CX(0, 1), Circuit(2))
    matcher = RuleMatcher([rule])

    mtch = matcher.find_match(c)

    c.apply_rewrite(mtch)

    out = c.to_tket1()

    assert out == Circuit(4).CX(0, 2)


def test_multiple_rules():
    circ = Tk2Circuit(Circuit(3).CX(0, 1).H(0).H(1).H(2).Z(0).H(0).H(1).H(2))

    rule1 = Rule(Circuit(1).H(0).Z(0).H(0), Circuit(1).X(0))
    rule2 = Rule(Circuit(1).H(0).H(0), Circuit(1))
    matcher = RuleMatcher([rule1, rule2])

    match_count = 0
    while match := matcher.find_match(circ):
        match_count += 1
        circ.apply_rewrite(match)

    assert match_count == 3

    out = circ.to_tket1()
    assert out == Circuit(3).CX(0, 1).X(0)



from enum import Enum

class HugrType:
    """Value types in HUGR."""

    def __init__(self, extension: str, type_name: str, bound: TypeBound) -> None:
        """Create a new named Custom type."""

    @staticmethod
    def qubit() -> HugrType:
        """Qubit type from HUGR prelude."""

    @staticmethod
    def bool() -> HugrType:
        """Boolean type (HUGR 2-ary unit sum)."""

class TypeBound(Enum):
    """HUGR type bounds."""

    Any = 0  # Any type
    Copyable = 1  # Copyable type
    Eq = 2  # Equality-comparable type



from tket_exts import (
    opaque_bool,
    debug,
    guppy,
    rotation,
    futures,
    qsystem,
    qsystem_random,
    qsystem_utils,
    quantum,
    result,
    wasm,
)


__all__ = [
    "debug",
    "opaque_bool",
    "guppy",
    "rotation",
    "futures",
    "qsystem",
    "qsystem_random",
    "qsystem_utils",
    "quantum",
    "result",
    "wasm",
]



from typing import Any, Callable
from pytket._tket.circuit import Circuit as Tk1Circuit

from tket._tket.ops import TketOp
from hugr.envelope import EnvelopeConfig

try:
    from warnings import deprecated  # type: ignore[attr-defined]
except ImportError:
    # Python < 3.13
    def deprecated(_msg):  # type: ignore[no-redef]
        def _deprecated(func):
            return func

        return _deprecated

class Tk2Circuit:
    """Rust representation of a TKET circuit."""

    def __init__(self, circ: Tk1Circuit) -> None:
        """Create a Tk2Circuit from a pytket Circuit."""

    def __hash__(self) -> int:
        """Compute the circuit hash by traversal."""

    def __copy__(self) -> Tk2Circuit:
        """Create a copy of the circuit."""

    def __deepcopy__(self) -> Tk2Circuit:
        """Create a deep copy of the circuit."""

    def hash(self) -> int:
        """Compute the circuit hash by traversal."""

    def circuit_cost(self, cost_fn: Callable[[TketOp], Any]) -> int:
        """Compute the cost of the circuit. Return value must implement __add__."""

    def num_operations(self) -> int:
        """The number of operations in the circuit.

        This includes [`TketOp`]s, pytket ops, and any other custom operations.

        Nested circuits are traversed to count their operations.
        """

    def node_op(self, node: Node) -> bytes:
        """If the node corresponds to a custom op, return it. Otherwise, raise an error."""

    def to_tket1(self) -> Tk1Circuit:
        """Convert to pytket Circuit."""

    def apply_rewrite(self, rw) -> None:
        """Apply a rewrite to the circuit."""

    def node_inputs(self, node: Node) -> list[Wire]:
        """The incoming wires to a node."""

    def node_outputs(self, node: Node) -> list[Wire]:
        """The outgoing wires from a node."""

    def input_node(self) -> Node:
        """The input node of the circuit."""

    def output_node(self) -> Node:
        """The output node of the circuit."""

    def to_hugr_json(self) -> str:
        """Encode the circuit as a HUGR json."""

    @staticmethod
    def from_hugr_json(json: str) -> Tk2Circuit:
        """Decode a HUGR json string to a Tk2Circuit."""

    def to_bytes(self, config: EnvelopeConfig) -> bytes:
        """Encode the circuit as a HUGR envelope, according to the given config.

        Some envelope formats can be encoded into a string. See :meth:`to_str`.

        Args:
            config: The envelope configuration to use.
        """

    def to_str(self, config: EnvelopeConfig | None = None) -> str:
        """Encode the circuit as a HUGR envelope string.

        Not all envelope formats can be encoded into a string.
        See :meth:`to_bytes` for a more general method.

        Args:
            config: The envelope configuration to use.
                If not given, uses the default textual encoding.
        """

    @staticmethod
    def from_bytes(envelope: bytes, function_name: str | None = None) -> Tk2Circuit:
        """Load a Circuit from a HUGR envelope.

        Some envelope formats can be read from a string. See :meth:`from_str`.

        Args:
            envelope: The byte string representing a Package.
            function_name: The name of the function in the envelope's module to load.
                Defaults to `main`.

        Returns:
            The loaded circuit.
        """

    @staticmethod
    def from_str(envelope: str, function_name: str | None = None) -> Tk2Circuit:
        """Load a Circuit from a HUGR envelope string.

        Not all envelope formats can be represented as strings.
        See :meth:`from_bytes` for a more general method.

        Args:
            envelope: The string representing a Package.
            function_name: The name of the function in the envelope's module to load.
                Defaults to `main`.

        Returns:
            The loaded circuit.
        """

    @deprecated("Use HUGR envelopes instead. See the `to_bytes` and `to_str` methods.")
    def to_package_json(self) -> str:
        """Encode the circuit as a HUGR Package json."""

    @staticmethod
    @deprecated(
        "Use HUGR envelopes instead. See the `from_bytes` and `from_str` methods."
    )
    def from_package_json(json: str, function_name: str | None = None) -> Tk2Circuit:
        """Decode a HUGR Package json to a circuit.

        Traverses the package's modules in order until it finds one containing a
        function named `function_name`, and loads it as a circuit.

        If the json is a hugr json, it will be decoded as a `main` function in an empty module.

        When `function_name` is not given, it defaults to `main`.
        """

    def to_tket1_json(
        self,
    ) -> str:
        """Encode the circuit as a pytket json string."""

    @staticmethod
    def from_tket1_json(json: str) -> Tk2Circuit:
        """Decode a pytket json string to a Tk2Circuit."""

class Node:
    """Handle to node in HUGR."""

    def __init__(self, idx: int) -> None:
        """Create a new node handle."""

class Wire:
    """An outgoing edge from a node in a HUGR, defined by the node and outgoing port."""

    def node(self) -> Node:
        """Source node of wire."""

    def port(self) -> int:
        """Source port of wire."""

class CircuitCost:
    """A cost function for circuits."""

    def __init__(self, cost: Any) -> None:
        """Create a new circuit cost.

        The cost object must implement __add__, __sub__, __eq__, and __lt__."""

def render_circuit_dot(hugr: Tk2Circuit | Tk1Circuit) -> str: ...
def render_circuit_mermaid(hugr: Tk2Circuit | Tk1Circuit) -> str: ...
def validate_circuit(hugr: Tk2Circuit | Tk1Circuit) -> None: ...

class HugrError(Exception): ...
class BuildError(Exception): ...
class ValidationError(Exception): ...
class HUGRSerializationError(Exception): ...
class TK1ConvertError(Exception): ...



from dataclasses import dataclass

from pytket._tket.circuit import Circuit

from tket.circuit import (
    Tk2Circuit,
    render_circuit_dot,
)
from tket.ops import TketOp


@dataclass
class CustomCost:
    gate_count: int
    h_count: int

    def __add__(self, other):
        return CustomCost(
            self.gate_count + other.gate_count, self.h_count + other.h_count
        )


def test_cost():
    circ = Tk2Circuit(Circuit(4).CX(0, 1).H(1).CX(1, 2).CX(0, 3).H(0))

    print(circ.circuit_cost(lambda op: int(op == TketOp.CX)))

    assert circ.circuit_cost(lambda op: int(op == TketOp.CX)) == 3
    assert circ.circuit_cost(lambda op: CustomCost(1, op == TketOp.H)) == CustomCost(
        5, 2
    )


def test_hash():
    circA = Tk2Circuit(Circuit(4).CX(0, 1).CX(1, 2).CX(0, 3))
    circB = Tk2Circuit(Circuit(4).CX(1, 2).CX(0, 1).CX(0, 3))
    circC = Tk2Circuit(Circuit(4).CX(0, 1).CX(0, 3).CX(1, 2))

    assert hash(circA) != hash(circB)
    assert hash(circA) == hash(circC)


def test_conversion():
    tk1 = Circuit(4).CX(0, 2).CX(1, 2).CX(1, 3)
    tk1_dot = render_circuit_dot(tk1)

    tk2 = Tk2Circuit(tk1)
    tk2_dot = render_circuit_dot(tk2)

    assert type(tk2) is Tk2Circuit
    assert tk1_dot == tk2_dot

    tk1_back = tk2.to_tket1()

    assert tk1_back == tk1
    assert type(tk1_back) is Circuit



from __future__ import annotations

from enum import Enum, auto
from typing import Protocol

import tket

from tket._tket.ops import CustomOp
from tket.types import QB_T

__all__ = ["CustomOp", "ToCustomOp", "TketOp", "Pauli"]


class ToCustomOp(Protocol):
    """Operation that can be converted to a HUGR CustomOp."""

    def to_custom(self) -> CustomOp:
        """Convert to a custom operation."""
        ...

    @property
    def name(self) -> str:
        """Name of the operation."""
        ...


class TketOp(Enum):
    """A Tket built-in operation.

    Implements the `ToCustomOp` protocol.
    """

    H = auto()
    CX = auto()
    CY = auto()
    CZ = auto()
    CRz = auto()
    T = auto()
    Tdg = auto()
    S = auto()
    Sdg = auto()
    X = auto()
    Y = auto()
    Z = auto()
    Rz = auto()
    Rx = auto()
    Ry = auto()
    Toffoli = auto()
    Measure = auto()
    MeasureFree = auto()
    QAlloc = auto()
    TryQAlloc = auto()
    QFree = auto()
    Reset = auto()
    V = auto()
    Vdg = auto()

    def to_custom(self) -> CustomOp:
        """Convert to a custom operation."""
        return self._to_rs().to_custom()

    def _to_rs(self) -> tket._tket.ops.TketOp:
        """Convert to the Rust-backed TketOp representation."""
        return tket._tket.ops.TketOp(self.name)

    @staticmethod
    def _from_rs(op: tket._tket.ops.TketOp) -> "TketOp":
        """Convert from the Rust-backed TketOp representation."""
        return TketOp[op.name]

    def __eq__(self, other: object) -> bool:
        """Check if two TketOps are equal."""
        if isinstance(other, TketOp):
            return self.name == other.name
        elif isinstance(other, tket._tket.ops.TketOp):
            return self == TketOp._from_rs(other)
        elif isinstance(other, str):
            return self.name == other
        return False


class Pauli(Enum):
    """Simple enum representation of Pauli matrices.

    Implements the `ToCustomOp` protocol.
    """

    I = auto()  # noqa: E741
    X = auto()
    Y = auto()
    Z = auto()

    def to_custom(self) -> CustomOp:
        extension_name = "tket.quantum"
        gate_name = self.name
        return CustomOp(extension_name, gate_name, [QB_T], [QB_T])

    def _to_rs(self) -> tket._tket.ops.Pauli:
        """Convert to the Rust-backed Pauli representation."""
        return tket._tket.ops.Pauli(self.name)

    @staticmethod
    def _from_rs(pauli: tket._tket.ops.Pauli) -> "Pauli":
        """Convert from the Rust-backed Pauli representation."""
        return Pauli[pauli.name]

    def __eq__(self, other: object) -> bool:
        """Check if two Paulis are equal."""
        if isinstance(other, Pauli):
            return self.name == other.name
        elif isinstance(other, tket._tket.ops.Pauli):
            return self == Pauli._from_rs(other)
        elif isinstance(other, str):
            return self.name == other
        return False



# Re-export native bindings
from ._tket.pattern import (
    Rule,
    RuleMatcher,
    CircuitPattern,
    PatternMatcher,
    PatternMatch,
    PatternID,
    InvalidPatternError,
    InvalidReplacementError,
)

__all__ = [
    "Rule",
    "RuleMatcher",
    "CircuitPattern",
    "PatternMatcher",
    "PatternMatch",
    "PatternID",
    "InvalidPatternError",
    "InvalidReplacementError",
]



from typing import TypeVar, Literal
from .circuit import Tk2Circuit
from pytket._tket.circuit import Circuit

from pathlib import Path

CircuitClass = TypeVar("CircuitClass", Circuit, Tk2Circuit)

class BadgerOptimiser:
    @staticmethod
    def load_precompiled(
        filename: Path, cost_fn: Literal["cx", "rz"] | None = None
    ) -> BadgerOptimiser:
        """
        Load a precompiled rewriter from a file.

        :param filename: The path to the file containing the precompiled rewriter.
        :param cost_fn: The cost function to use.
        """

    @staticmethod
    def compile_eccs(
        filename: Path, cost_fn: Literal["cx", "rz"] | None = None
    ) -> BadgerOptimiser:
        """
        Compile a set of ECCs and create a new rewriter.

        :param filename: The path to the file containing the ECCs.
        :param cost_fn: The cost function to use.
        """

    def optimise(
        self,
        circ: CircuitClass,
        timeout: int | None = None,
        progress_timeout: int | None = None,
        n_threads: int | None = None,
        split_circ: bool = False,
        queue_size: int | None = None,
        log_progress: Path | None = None,
    ) -> CircuitClass:
        """Optimise a circuit.

        :param circ: The circuit to optimise.
        :param timeout: Maximum time to spend on the optimisation.
        :param progress_timeout: Maximum time to wait between new best results.
        :param n_threads: Number of threads to use.
        :param split_circ: Split the circuit into subcircuits and optimise them separately.
        :param queue_size: Maximum number of circuits to keep in the queue of candidates.
        :param log_progress: Log progress to a CSV file.
        """



from enum import Enum
from typing import Any, Iterable

from tket._tket.types import HugrType

class TketOp(Enum):
    """A rust-backed Tket built-in operation."""

    def __init__(self, op: str) -> None:
        """Create a new TketOp from a string name."""

    @staticmethod
    def values() -> Iterable[TketOp]:
        """Iterate over all operation variants."""

    def to_custom(self) -> CustomOp:
        """Convert to a custom operation."""

    @property
    def name(self) -> str:
        """Get the string name of the operation."""

    @property
    def qualified_name(self) -> str:
        """Return the fully qualified name of the operation, including the extension."""

    def __str__(self) -> str:
        """Get the string name of the operation."""

    def __eq__(self, value: object) -> bool: ...

class Pauli(Enum):
    """Simple enum representation of Pauli matrices."""

    def __init__(self, pauli: str) -> None:
        """Create a new Pauli from a string name."""

    @property
    def name(self) -> str:
        """Get the string name of the Pauli."""

    @staticmethod
    def values() -> Iterable[Pauli]:
        """Iterate over all Pauli matrix variants."""

    @staticmethod
    def I() -> Pauli:  # noqa: E743
        """Identity Pauli matrix."""

    @staticmethod
    def X() -> Pauli:
        """Pauli-X matrix."""

    @staticmethod
    def Y() -> Pauli:
        """Pauli-Y matrix."""

    @staticmethod
    def Z() -> Pauli:
        """Pauli-Z matrix."""

    def __str__(self) -> str:
        """Get the string name of the Pauli."""

    def __eq__(self, value: Any) -> bool: ...

class CustomOp:
    """A HUGR custom operation."""

    def __init__(
        self,
        extension: str,
        op_name: str,
        input_types: list[HugrType],
        output_types: list[HugrType],
    ) -> None:
        """Create a new custom operation from name and input/output types."""

    def to_custom(self) -> CustomOp:
        """Convert to a custom operation. Identity operation."""

    @property
    def name(self) -> str:
        """Fully qualified (including extension) name of the operation."""



"""Some utility functions for the example notebooks."""

from hugr import Hugr
from hugr.envelope import EnvelopeConfig
from tket.passes import lower_to_pytket
from tket.circuit import Tk2Circuit
from typing import Any


def setup_jupyter_rendering():
    """Set up hugr rendering for Jupyter notebooks."""

    def _repr_hugr(
        h: Hugr, include=None, exclude=None, **kwargs
    ) -> dict[str, bytes | str]:
        return h.render_dot()._repr_mimebundle_(include, exclude, **kwargs)

    def _repr_tk2circ(
        circ: Tk2Circuit, include=None, exclude=None, **kwargs
    ) -> dict[str, bytes | str]:
        h = Hugr.load_json(circ.to_hugr_json())
        return _repr_hugr(h, include, exclude, **kwargs)

    setattr(Hugr, "_repr_mimebundle_", _repr_hugr)
    setattr(Tk2Circuit, "_repr_mimebundle_", _repr_tk2circ)


# TODO: Should this be part of the guppy API? Or tket?
# Takes a RawFunctionDef and converts it to a Tk2Circuit
def guppy_to_circuit(func_def: Any) -> Tk2Circuit:
    """Convert a Guppy function definition to a `Tk2Circuit`."""
    module = func_def.id.module
    assert module is not None, "Function definition must belong to a module"

    pkg = module.compile()

    json = pkg.package.to_str(EnvelopeConfig.TEXT)
    circ = Tk2Circuit.from_str(json, func_def.name)

    return lower_to_pytket(circ)



from tket.extensions import opaque_bool  # noqa: F401
from tket.extensions import debug  # noqa: F401
from tket.extensions import rotation  # noqa: F401
from tket.extensions import guppy  # noqa: F401
from tket.extensions import futures  # noqa: F401
from tket.extensions import qsystem  # noqa: F401
from tket.extensions import qsystem_random  # noqa: F401
from tket.extensions import qsystem_utils  # noqa: F401
from tket.extensions import quantum  # noqa: F401
from tket.extensions import result  # noqa: F401
from tket.extensions import wasm  # noqa: F401



from pathlib import Path
from typing import TypeVar

from .optimiser import BadgerOptimiser
from .circuit import Tk2Circuit
from pytket._tket.circuit import Circuit

CircuitClass = TypeVar("CircuitClass", Circuit, Tk2Circuit)

class CircuitChunks:
    def reassemble(self) -> Circuit | Tk2Circuit:
        """Reassemble the circuit from its chunks."""

    def circuits(self) -> list[Circuit | Tk2Circuit]:
        """Returns clones of the split circuits."""

    def update_circuit(self, index: int, circ: Circuit | Tk2Circuit) -> None:
        """Replace a circuit chunk with a new version."""

class PullForwardError(Exception):
    """Error from a `PullForward` operation."""

def greedy_depth_reduce(circ: CircuitClass) -> tuple[CircuitClass, int]:
    """Greedy depth reduction of a circuit.

    Returns the reduced circuit and the depth reduction.
    """

def lower_to_pytket(circ: CircuitClass) -> CircuitClass:
    """Lower the high-level operations in a Hugr so it can be interpreted by pytket."""

def badger_optimise(
    circ: CircuitClass,
    optimiser: BadgerOptimiser,
    max_threads: int | None = None,
    timeout: int | None = None,
    progress_timeout: int | None = None,
    max_circuit_count: int | None = None,
    log_dir: Path | None = None,
    rebase: bool | None = False,
) -> CircuitClass:
    """Optimise a circuit using the Badger optimiser.

    HyperTKET's best attempt at optimising a circuit using circuit rewriting
    and the given Badger optimiser.

    By default, the input circuit will be rebased to Nam, i.e. CX + Rz + H before
    optimising. This can be deactivated by setting `rebase` to `false`, in which
    case the circuit is expected to be in the Nam gate set.

    Will use at most `max_threads` threads (plus a constant). Defaults to the
    number of CPUs available.

    The optimisation will terminate at the first of the following timeout
    criteria, if set:
    - `timeout` seconds (default: 15min) have elapsed since the start of the
      optimisation
    - `progress_timeout` (default: None) seconds have elapsed since progress
      in the cost function was last made
    - `max_circuit_count` (default: None) circuits have been explored.

    Log files will be written to the directory `log_dir` if specified.
    """

def chunks(c: Circuit | Tk2Circuit, max_chunk_size: int) -> CircuitChunks:
    """Split a circuit into chunks of at most `max_chunk_size` gates."""



# Re-export native bindings

from .._tket.circuit import (
    Tk2Circuit,
    Node,
    Wire,
    CircuitCost,
    validate_circuit,
    render_circuit_dot,
    render_circuit_mermaid,
    HugrError,
    BuildError,
    ValidationError,
    HUGRSerializationError,
    TK1ConvertError,
)

from .build import CircBuild, Command

__all__ = [
    "CircBuild",
    "Command",
    # Bindings.
    # TODO: Wrap these in Python classes.
    "Tk2Circuit",
    "Node",
    "Wire",
    "CircuitCost",
    "validate_circuit",
    "render_circuit_dot",
    "render_circuit_mermaid",
    "HugrError",
    "BuildError",
    "ValidationError",
    "HUGRSerializationError",
    "TK1ConvertError",
]



import tket
from tket.ops import TketOp, Pauli


def test_ops_roundtrip():
    for op in TketOp:
        assert TketOp._from_rs(op._to_rs()) == op

    for op in tket._tket.ops.TketOp.values():
        assert TketOp._from_rs(op)._to_rs() == op


def test_pauli_roundtrip():
    for pauli in Pauli:
        assert Pauli._from_rs(pauli._to_rs()) == pauli

    for pauli in tket._tket.ops.Pauli.values():
        assert Pauli._from_rs(pauli)._to_rs() == pauli



from __future__ import annotations
from typing import Iterable

from hugr import Hugr, tys, ops
from hugr.package import Package
from hugr.ext import Extension
from hugr.ops import ComWire, Command
from hugr.std.float import FLOAT_T
from hugr.build.function import Module
from hugr.build.tracked_dfg import TrackedDfg
from hugr.envelope import EnvelopeConfig
from tket.circuit import Tk2Circuit

from dataclasses import dataclass


class CircBuild(TrackedDfg):
    """Helper class to build a circuit from commands by tracking qubits,
    allowing commands to be specified by qubit index."""

    @classmethod
    def with_nqb(cls, n_qb: int) -> CircBuild:
        return cls(*[tys.Qubit] * n_qb, track_inputs=True)

    def finish_hugr(self) -> Hugr:
        """Finish building the hugr by setting all the qubits as the output.

        Returns:
            The finished Hugr.
        """
        return self.hugr

    def finish_package(
        self,
        *,
        other_extensions: Iterable[Extension] | None = None,
        function_name="main",
    ) -> Package:
        """Finish building the package by setting all the qubits as the output
        and wrap it in a hugr package with the required extensions.

        Args:
            other_extensions: Other extensions to include in the package.
            function_name: The name of the function containing the circuit in
                the package's module. Defaults to "main".
        Returns:
            The finished package.
        """
        # TODO: Replace with `finish_hugr` once extensions are included in the hugr itself.
        # See https://github.com/CQCL/hugr/pull/1621
        import tket.extensions as ext

        extensions = [
            ext.rotation(),
            ext.futures(),
            ext.qsystem(),
            ext.quantum(),
            ext.result(),
            *(other_extensions or []),
        ]

        # Convert the DFG into a Function definition
        dfg_op = self.hugr[self.hugr.entrypoint].op
        assert type(dfg_op) is ops.DFG, "CircBuild must have a Dfg root"
        self.hugr[self.hugr.entrypoint].op = ops.FuncDefn(
            function_name, inputs=dfg_op.inputs, _outputs=dfg_op.outputs
        )

        # Insert it into a module, as required by the package.
        module = Module()
        module.hugr.insert_hugr(self.hugr)

        return Package(modules=[module.hugr], extensions=extensions)

    def finish(self, other_extensions: list[Extension] | None = None) -> Tk2Circuit:
        """Finish building the circuit by setting all the qubits as the output
        and validate."""

        return Tk2Circuit.from_str(
            self.finish_package(
                other_extensions=other_extensions, function_name="main"
            ).to_str(EnvelopeConfig.TEXT),
            "main",
        )


def from_coms(*args: Command) -> Tk2Circuit:
    """Build a circuit from a sequence of commands, assuming
    only qubits are referred to by index."""
    commands: list[Command] = []
    n_qb = 0
    # traverses commands twice which isn't great
    for arg in args:
        max_qb = max(i for i in arg.incoming if isinstance(i, int)) + 1
        n_qb = max(n_qb, max_qb)
        commands.append(arg)

    build = CircBuild.with_nqb(n_qb)
    build.extend(*commands)
    build.set_tracked_outputs()
    return build.finish()


def load_custom(serialized: bytes) -> ops.Custom:
    import hugr._serialization.ops as sops
    import json

    # TODO: We should return an "ExtOp" instead
    ext = json.loads(serialized)
    return sops.ExtensionOp(**ext).deserialize()


def id_circ(n_qb: int) -> Tk2Circuit:
    b = CircBuild.with_nqb(n_qb)
    b.set_tracked_outputs()
    return b.finish()


@dataclass(frozen=True)
class QuantumOps(ops.Custom):
    extension: tys.ExtensionId = "tket.quantum"


_OneQbSig = tys.FunctionType.endo([tys.Qubit])


@dataclass(frozen=True)
class OneQbGate(QuantumOps):
    op_name: str  # type: ignore[misc] # no-default fields follows one with a default
    num_out: int = 1
    signature: tys.FunctionType = _OneQbSig

    def __call__(self, q: ComWire) -> Command:
        return super().__call__(q)


H = OneQbGate("H")
PauliX = OneQbGate("X")
PauliY = OneQbGate("Y")
PauliZ = OneQbGate("Z")

_TwoQbSig = tys.FunctionType.endo([tys.Qubit] * 2)


@dataclass(frozen=True)
class TwoQbGate(QuantumOps):
    op_name: str  # type: ignore[misc] # no-default fields follows one with a default
    num_out: int = 2
    signature: tys.FunctionType = _TwoQbSig

    def __call__(self, q0: ComWire, q1: ComWire) -> Command:
        return super().__call__(q0, q1)


CX = TwoQbGate("CX")

_MeasSig = tys.FunctionType([tys.Qubit], [tys.Qubit, tys.Bool])


@dataclass(frozen=True)
class MeasureDef(QuantumOps):
    op_name: str = "Measure"
    num_out: int = 2
    signature: tys.FunctionType = _MeasSig

    def __call__(self, q: ComWire) -> Command:
        return super().__call__(q)


Measure = MeasureDef()

# TODO use angle type once extension is serialised.
_RzSig = tys.FunctionType([tys.Qubit, FLOAT_T], [tys.Qubit])


@dataclass(frozen=True)
class RzDef(QuantumOps):
    op_name: str = "Rz"
    num_out: int = 1
    signature: tys.FunctionType = _RzSig

    def __call__(self, q: ComWire, fl_wire: ComWire) -> Command:
        return super().__call__(q, fl_wire)


Rz = RzDef()


_QallocSig = tys.FunctionType([], [tys.Qubit])


@dataclass(frozen=True)
class QAllocDef(QuantumOps):
    op_name: str = "QAlloc"
    num_out: int = 1
    signature: tys.FunctionType = _QallocSig

    def __call__(self) -> Command:
        return super().__call__()


QAlloc = QAllocDef()


_QfreeSig = tys.FunctionType([tys.Qubit], [])


@dataclass(frozen=True)
class QFreeDef(QuantumOps):
    op_name: str = "QFree"
    num_out: int = 0
    signature: tys.FunctionType = _QfreeSig

    def __call__(self, qb: ComWire) -> Command:
        return super().__call__(qb)


QFree = QFreeDef()



from pytket import Circuit
from pytket.qasm import circuit_from_qasm
from tket.pattern import CircuitPattern, PatternMatcher


def test_simple_matching():
    """a simple circuit matching test"""
    c = Circuit(2).CX(0, 1).H(1).CX(0, 1)

    p1 = CircuitPattern(Circuit(2).CX(0, 1).H(1))
    p2 = CircuitPattern(Circuit(2).H(0).CX(1, 0))

    matcher = PatternMatcher(iter([p1, p2]))

    assert len(matcher.find_matches(c)) == 2


def test_non_convex_pattern():
    """two-qubit circuits can't match three-qb ones"""
    p1 = CircuitPattern(Circuit(3).CX(0, 1).CX(1, 2))
    matcher = PatternMatcher(iter([p1]))

    c = Circuit(2).CX(0, 1).CX(1, 0)
    assert len(matcher.find_matches(c)) == 0

    c = Circuit(3).CX(0, 1).CX(1, 0).CX(1, 2)
    assert len(matcher.find_matches(c)) == 0

    c = Circuit(3).H(0).CX(0, 1).CX(1, 0).CX(0, 2)
    assert len(matcher.find_matches(c)) == 1


def test_larger_matching():
    """a larger crafted circuit with matches WIP"""
    c = circuit_from_qasm("test_files/simple.qasm")

    p1 = CircuitPattern(Circuit(2).CX(0, 1).H(1))
    p2 = CircuitPattern(Circuit(2).H(0).CX(1, 0))
    p3 = CircuitPattern(Circuit(2).CX(0, 1).CX(1, 0))
    p4 = CircuitPattern(Circuit(3).CX(0, 1).CX(1, 2))

    matcher = PatternMatcher(iter([p1, p2, p3, p4]))

    assert len(matcher.find_matches(c)) == 6



from typing import Iterator
from .circuit import Node, Tk2Circuit
from .rewrite import CircuitRewrite
from pytket._tket.circuit import Circuit as Tk1Circuit

class Rule:
    """A rewrite rule defined by a left hand side and right hand side of an equation."""

    def __init__(
        self,
        l: Tk1Circuit | Tk2Circuit,  # noqa: E741
        r: Tk1Circuit | Tk2Circuit,
    ) -> None:
        """Create a new rewrite rule."""

    def lhs(self) -> Tk2Circuit:
        """Get the left hand side of the rule.

        This is the pattern that is matched in the circuit.
        """

    def rhs(self) -> Tk2Circuit:
        """Get the right hand side of the rule.

        This is the pattern that is replaced in the circuit.
        """

class RuleMatcher:
    """A matcher for multiple rewrite rule."""

    def __init__(self, rules: list[Rule]) -> None:
        """Create a new rule matcher."""

    def find_match(self, circ: Tk2Circuit) -> CircuitRewrite | None:
        """Find a match of the rules in the circuit."""

    def find_matches(self, circ: Tk2Circuit) -> list[CircuitRewrite]:
        """Find all matches of the rules in the circuit."""

class CircuitPattern:
    """A pattern that matches a circuit exactly."""

    def __init__(self, circ: Tk1Circuit | Tk2Circuit) -> None:
        """Create a new circuit pattern."""

class PatternMatcher:
    """A matcher object for fast pattern matching on circuits."""

    def __init__(self, patterns: Iterator[CircuitPattern]) -> None:
        """Create a new pattern matcher."""

    def find_match(self, circ: Tk2Circuit) -> PatternMatch | None:
        """Find a match of the patterns in the circuit."""

    def find_matches(self, circ: Tk2Circuit) -> list[PatternMatch]:
        """Find all matches of the patterns in the circuit."""

class PatternMatch:
    """A convex pattern match in a circuit"""

    def pattern_id(self) -> PatternID:
        """The id of the matched pattern."""

    def root(self) -> Node:
        """The root node for the pattern in the matched circuit."""

class PatternID:
    """An identifier for a pattern in a pattern matcher."""

    def __int__(self) -> int:
        """Get the integer value of the pattern id."""

class InvalidPatternError(Exception):
    """Conversion error between a pattern and a circuit."""

class InvalidReplacementError(Exception):
    """An error occurred while constructing a pattern match replacement."""



# Re-export native bindings
from ._tket.rewrite import ECCRewriter, CircuitRewrite, Subcircuit

__all__ = [
    "default_ecc_rewriter",
    # Bindings.
    # TODO: Wrap these in Python classes.
    "ECCRewriter",
    "CircuitRewrite",
    "Subcircuit",
]


def default_ecc_rewriter() -> ECCRewriter:
    """Load the default ecc rewriter."""
    try:
        import tket_eccs
    except ImportError:
        raise ValueError(
            "The default rewriter is not available. Please specify a path to a rewriter or install tket-eccs."
        )

    rewriter = tket_eccs.nam_6_3()
    return ECCRewriter.load_precompiled(rewriter)



import itertools

import pytest
from pytket._tket.circuit import Circuit

from hugr.ops import Custom
from hugr.hugr import Wire
from tket.circuit import (
    Tk2Circuit,
    Node as Tk2Node,
)
from tket.circuit.build import (
    CircBuild,
    H,
    from_coms,
    CX,
    PauliX,
    PauliY,
    PauliZ,
    Measure,
    QAlloc,
    QFree,
    load_custom,
    id_circ,
)
from hugr.std.logic import Not
from tket.pattern import Rule, RuleMatcher  # type: ignore
from tket.rewrite import CircuitRewrite, Subcircuit  # type: ignore


@pytest.fixture
def merge_rules() -> list[Rule]:
    paulis = [PauliX(0), PauliY(0), PauliZ(0)]
    identities = [Rule(from_coms(p, p), id_circ(1)) for p in paulis]

    off_diag = [
        Rule(from_coms(p0, p1), from_coms(p2))
        for p0, p1, p2 in itertools.permutations(paulis)
    ]
    return [*identities, *off_diag]


@pytest.fixture
def propagate_rules() -> list[Rule]:
    hadamard_rules = [
        Rule(from_coms(PauliX(0), H(0)), from_coms(H(0), PauliZ(0))),
        Rule(from_coms(PauliZ(0), H(0)), from_coms(H(0), PauliX(0))),
    ]

    cx_rules = [
        Rule(from_coms(PauliZ(0), CX(0, 1)), from_coms(CX(0, 1), PauliZ(0))),
        Rule(from_coms(PauliX(1), CX(0, 1)), from_coms(CX(0, 1), PauliX(1))),
        Rule(from_coms(PauliZ(1), CX(0, 1)), from_coms(CX(0, 1), PauliZ(0), PauliZ(1))),
        Rule(from_coms(PauliX(0), CX(0, 1)), from_coms(CX(0, 1), PauliX(0), PauliX(1))),
    ]

    return [*hadamard_rules, *cx_rules]


@pytest.fixture
def measure_rules() -> list[Rule]:
    # X flips the measured bit
    r_build = CircBuild.with_nqb(1)
    (_, measure) = r_build.extend(PauliX(0), Measure(0))
    r_build.set_indexed_outputs(0, measure[1])
    ltk = r_build.finish()

    r_build = CircBuild.with_nqb(1)
    b = r_build.add(Measure(0))[1]
    b = r_build.add(Not(b))[0]
    r_build.set_indexed_outputs(0, b)
    rtk = r_build.finish()

    rules = [Rule(ltk, rtk)]

    # Z does not affect measurement result
    r_build = CircBuild.with_nqb(1)
    (_, measure) = r_build.extend(PauliZ(0), Measure(0))
    r_build.set_indexed_outputs(0, measure[1])
    ltk = r_build.finish()

    r_build = CircBuild.with_nqb(1)
    b = r_build.add(Measure(0))[1]
    r_build.set_indexed_outputs(0, b)
    rtk = r_build.finish()

    rules.append(Rule(ltk, rtk))

    return rules


@pytest.fixture
def propagate_matcher(
    merge_rules: list[Rule], propagate_rules: list[Rule], measure_rules: list[Rule]
) -> RuleMatcher:
    # TODO: This broke with the update to hugr 0.8.0.
    # Custom ops in hugrs must be resolved into `ExtOp`s before they can be used.
    # In this case, we need to define an extension for PauliOps, and make sure the
    # `Rule`s use them, otherwise `RuleMatcher` will complain about finding opaque ops.
    return RuleMatcher([*merge_rules, *propagate_rules, *measure_rules])


def apply_exhaustive(circ: Tk2Circuit, matcher: RuleMatcher) -> int:
    """Apply the first matching rule until no more matches are found. Return the
    number of rewrites applied."""
    match_count = 0
    while match := matcher.find_match(circ):
        match_count += 1
        circ.apply_rewrite(match)

    return match_count


def add_error_after(circ: Tk2Circuit, wire: Wire, error: Custom):
    """Use a rewrite to insert an operation on a qubit wire assuming the error is
    a one qubit operation, and the source gate of the wire only acts on qubits."""
    port = wire.out_port()
    node = Tk2Node(port.node.idx)
    port_offset = port.offset
    n_qb = len(circ.node_outputs(node)) - 1  # ignore Order port
    subc = Subcircuit([node], circ)
    replace_build = CircBuild.with_nqb(n_qb)
    current = load_custom(circ.node_op(node))
    replace_build.add(current(*list(range(n_qb))))
    replace_build.add(error(port_offset))
    replace_build.set_tracked_outputs()
    replacement = replace_build.finish()

    rw = CircuitRewrite(subc, circ, replacement)

    circ.apply_rewrite(rw)


def final_pauli_string(circ: Tk2Circuit) -> str:
    """Assuming the circuit only has qubit outputs - check the final operations
    on each qubit, and if they are paulis concatenate them into a string."""

    def map_op(op: Custom) -> str:
        n = op.name()
        return n if n in ("X", "Y", "Z") else "I"

    # TODO ignore non-qubit outputs
    return "".join(
        map_op(load_custom(circ.node_op(w.node())))
        for w in circ.node_inputs(circ.output_node())
    )


@pytest.mark.skip(reason="Broken with hugr 0.8.0. See comment in `propagate_matcher`.")
def test_simple_z_prop(propagate_matcher: RuleMatcher):
    c = CircBuild.with_nqb(2)

    (h_node_e, *_) = c.extend(H(0), H(0), CX(0, 1))
    c.set_tracked_outputs()

    t2c = c.finish()

    add_error_after(t2c, h_node_e[0], PauliX)

    assert t2c.to_tket1() == Circuit(2).H(0).X(0).H(0).CX(0, 1)

    assert apply_exhaustive(t2c, propagate_matcher) == 2

    assert t2c.to_tket1() == Circuit(2).H(0).H(0).CX(0, 1).Z(0)

    assert final_pauli_string(t2c) == "ZI"


@pytest.mark.skip(reason="Broken with hugr 0.8.0. See comment in `propagate_matcher`.")
def test_cat(propagate_matcher: RuleMatcher):
    c = CircBuild.with_nqb(4)
    (h_node, *_) = c.extend(
        H(2),
        CX(2, 1),
        CX(2, 3),
        CX(1, 0),
    )
    c.set_tracked_outputs()
    t2c = c.finish()

    add_error_after(t2c, h_node[0], PauliX)
    assert t2c.to_tket1() == Circuit(4).H(2).X(2).CX(2, 1).CX(2, 3).CX(1, 0)

    assert apply_exhaustive(t2c, propagate_matcher) == 3

    assert t2c.to_tket1() == Circuit(4).H(2).CX(2, 1).CX(2, 3).CX(1, 0).X(0).X(1).X(
        2
    ).X(3)

    assert final_pauli_string(t2c) == "XXXX"


def test_alloc_free():
    c = CircBuild()
    alloc = c.add(QAlloc())
    c.add(QFree(alloc))
    c.set_outputs()
    c.finish()  # validates


@pytest.mark.skip(reason="Broken with hugr 0.8.0. See comment in `propagate_matcher`.")
def test_measure(propagate_matcher: RuleMatcher):
    c = CircBuild.with_nqb(2)
    (_, m0, m1) = c.extend(PauliX(0), Measure(0), Measure(1))

    c.set_indexed_outputs(0, m0[1], 1, m1[1])
    before = c.finish()
    """
    ──►X───►Measure─►
            │
            └────►
    ───────►Measure─►
            │
            └────►
    """
    assert apply_exhaustive(before, propagate_matcher) == 1
    c = CircBuild.with_nqb(2)
    (m0, m1) = c.extend(Measure(0), Measure(1))
    nt = c.add(Not(m0[1]))
    c.set_indexed_outputs(0, nt, 1, m1[1])
    after = c.finish()
    """
    ──►Measure──────►
            └─►Not─►
    ──────►Measure──►
                └────►
    """

    # can't compare using tket1 circuits because measure can't be converted.
    assert hash(before) == hash(after)



from pathlib import Path
from typing import Optional, Literal

from pytket import Circuit
from pytket.passes import CustomPass, BasePass

from tket import optimiser

# Re-export native bindings
from ._tket.passes import (
    CircuitChunks,
    greedy_depth_reduce,
    lower_to_pytket,
    badger_optimise,
    chunks,
    PullForwardError,
)

__all__ = [
    "badger_pass",
    # Bindings.
    # TODO: Wrap these in Python classes.
    "CircuitChunks",
    "greedy_depth_reduce",
    "lower_to_pytket",
    "badger_optimise",
    "chunks",
    "PullForwardError",
]


def badger_pass(
    rewriter: Optional[Path] = None,
    max_threads: Optional[int] = None,
    timeout: Optional[int] = None,
    progress_timeout: Optional[int] = None,
    max_circuit_count: Optional[int] = None,
    log_dir: Optional[Path] = None,
    rebase: bool = False,
    cost_fn: Literal["cx", "rz"] | None = None,
) -> BasePass:
    """Construct a Badger pass.

    The Badger optimiser requires a pre-compiled rewriter produced by the
    `compile-rewriter <https://github.com/CQCL/tket2/tree/main/badger-optimiser>`_
    utility. If `rewriter` is not specified, a default one will be used.

    The cost function to minimise can be specified by passing `cost_fn` as `'cx'`
    or `'rz'`. If not specified, the default is `'cx'`.

    The arguments `max_threads`, `timeout`, `progress_timeout`, `max_circuit_count`,
    `log_dir` and `rebase` are optional and will be passed on to the Badger
    optimiser if provided."""
    if rewriter is None:
        try:
            import tket_eccs
        except ImportError:
            raise ValueError(
                "The default rewriter is not available. Please specify a path to a rewriter or install tket-eccs."
            )

        rewriter = tket_eccs.nam_6_3()
    opt = optimiser.BadgerOptimiser.load_precompiled(rewriter, cost_fn=cost_fn)

    def apply(circuit: Circuit) -> Circuit:
        """Apply Badger optimisation to the circuit."""
        return badger_optimise(
            circuit,
            optimiser=opt,
            max_threads=max_threads,
            timeout=timeout,
            progress_timeout=progress_timeout,
            max_circuit_count=max_circuit_count,
            log_dir=log_dir,
            rebase=rebase,
        )

    return CustomPass(apply, label="tket.badger_pass")
